function [x,RRE]=l23Reg(A,b,mu,maxit,tol,x_true)
%This function implements the thresholding algorithm for the solution of
%the minimization problem
%
% argmin 1/2||A*x-b||_2^2+3*mu/2||x||_{2/3}^{2/3}.
%
% The sintax of the function is
% [x,RRE]=l23Reg(A,b,mu,maxit,tol,x_true);
%
% INPUT:
% A: system matrix (it can be an object, it requires that the operations
%                   A*x, A'*x, and normest, are defined);
% b: data vector;
% mu: regularization parameter;
% maxit: maximum  number of iterations;
% tol: tolerance of the stopping criterion;
% x_true: exact solution (optional, if empty or not given the output
%         argument RRE is not populated).
%
% OUTPUT:
% x: approximate solution;
% RRE: vector containing the Relative Restoration Error at each iteration
%      (populated only if x_true is non empty).
%
% This function is part of the lplq package described in the paper:
% A.Buccini and L.Reichel, Software for limited memory restarted
% $\ell^p-\ell^q$ minimization methods using generalized Krylov subspaces, 
% Electron. Trans. Numer. Anal., Submitted(2024), pp. 1--24.

% Check if x_true is provided
if nargin==5
    x_true=[];
end

% Preallocate memory for RRE
if nargout==2 && ~isempty(x_true)
    RRE=zeros(maxit+1,1);
    RRE(1)=1;
    nrmx=norm(x_true);
else
    RRE=[];
end

% Initial guess
x=zeros(size(A'*b));
% Set steplength
eta=0.9/normest(A)^2;
% Begin iterations
for k=1:maxit
    % Store previous iteration for the stopping criterion
    xold=x; 
    % Compute the residual
    r=b-A*x;
    % Compute a step of gradient descent
    x=x+eta*(A'*r);
    % Threshold the solution
    x=thresh23(x,mu*eta);
    % If RRE is required and x_true is given and non-empty compute the RRE
    if nargout==2 && ~isempty(x_true)
        RRE(k+1)=norm(x-x_true)/nrmx;
    end
    % Check the stopping criterion
    if norm(x-xold)<tol*norm(xold)
        if nargout==2 && ~isempty(x_true)
            RRE=RRE(1:k+1);
        end
        break
    end
end
end

function y=thresh23(z,mu)
%l^2/3 thresholding
t=(48^(0.25))/3*mu^(3/4);
phi=@(lambda,x) 2/sqrt(3)*lambda^0.25*sqrt(cosh(acosh(27/16*lambda^(-3/2)*x.^2)/3));
f=@(lambda,x) ((abs(phi(lambda,x))+sqrt(2*abs(x)./abs(phi(lambda,x))-abs(phi(lambda,x)).^2))/2).^3.*sign(x);
y=z;
y(abs(y)<=t)=0;
y(abs(y)>t)=f(mu,y(abs(y)>t));
end