function [x,outInfo]=lplq(A,b,options)
% This program solves the minimization problem
% x=argmin 1/p*||Ax-b||_p^p+mu/q*||Lx||_q^q
%
% The sintax of the function is
% [x,outInfo]=lplq(A,b,options);
%
% INPUT:
% A: system matrix (it can be an object, it requires that the operations
%                   A*x and A'*x are defined);
% b: data vector;
% options: struture containing the parameters of the algorithm
%        .p: value of p 0<p<=2 (default value p=2);
%        .q: value of q 0<q<=2 (default value q=0.1);
%        .L: regularization matrix (it can be an object, it requires that
%                                   the operations L*x and L'*x are defined
%                                   default value L=I);
%        .mu: regularization paramter. If it is populated, then
%             this value of mu is used. If it is not populated, an
%             automatic rule choice (in the field options.choiceRule) is
%             used (default value mu=[]). Both fields .mu and .choiceRule
%             cannot be empty at the same time;
%        .initialGuess: initial guess for the iterations. It can either be
%                       a vector or empty. In the latter case
%                       the vector A'*b is used as an initial guess
%                       (default value []);
%        .choiceRule: automatic rule for the computation of the
%                     regularization paramter (default value choiceRule='GCV').
%                     The possible values are:
%                     - 'DP': discrepancy principle (if selected the field
%                             options.noiseNorm cannot be empty, the
%                             majorantType field is automatically set to
%                             'fixed', and p is fixed to 2);
%                     - 'GCV': generalized cross validation (if selected the
%                              majorantType field is automatically set to
%                              'adaptive'). If p<=1, then the smoothed
%                              version is used;
%                     - 'RWP': residual whiteness principle (available only
%                              for p=2 or p=1);
%                     - 'CV': cross validation;
%                     - 'MCV': modified cross validation;
%                     - []: the value in options.mu is used. In this case
%                           options.mu cannot be empty.
%        .majorantType: type of majorant employed. The possible values are
%                       'fixed' and 'adaptive' (default value 'adaptive');
%        .epsilon: smoothing parameter for the algorithm (default value
%                  epsilon=1e-3);
%        .tol: tollerance for the stopping criterion (default value=1e-4);
%        .maxIt: maximum number of iterations (default value=100);
%        .noiseNorm: norm of the noise that corrupts the data. This field
%                    cannot be empty if choiceRule='DP' (default value=[]);
%        .tau: parameter for the discrepancy principle (default
%              value=1.01);
%        .size: size of the data in case b is a vectorized version of a 
%               multi-dimensional vector. This is used for the choiceRule 
%               'RWP' and 'GCV' if p<=1. The vector sizes must have at most 
%               three entries if the GCV is employed. If it is empty b is 
%               assumed to b one-dimensional (default values=[]);
%        .trainingPercent: percentage of data used to determine the
%                          regularization parameter in the CV and MCV
%                          (default value=90);
%        .trainingTestNum: number of different tests performed for the CV
%                          and MCV (default value=10);
%        .trainingMu: set of regularization parameters used in the CV and
%                     MCV (default logspace(-3,1,10));
%        .restart: number of iterations after which the generalized Krylov
%                  subspace is restarted. If empty or not given,
%                  options.restart=options.maxIt+1;
%        .waitbar: if 'off' (default value), no waitbar is shown. If 'on' a
%                  waitbar is shown;
%        .xTrue: exact solution of the problem. If given, the field .RRE is
%                populated in the outInfo structure (default value=[]).
%
%
%
% OUTPUT:
% x: regularized solution
% outInfo: structure containing all the informations on the iterations
%         .iter: number of iterations performed. If choiceRule is either
%                'CV' or 'MCV' this is a matrix containing the result of
%                each test;
%         .mu: vector of the values of mu determined by the algorithm if
%              choiceRule is 'DP', 'GCV', or 'RWP';
%         .res: vector of the norms of the residual throughout the
%               iterations;
%         .RRE: vector for the relative restoration errors throughout the
%               iterations (populated only if .xTrue is not empty).
%
% use options=lplq('defaults') to obtain a structure containing all the
% default values.
%
% This function is part of the lplq package described in the paper:
% A.Buccini and L.Reichel, Software for limited memory restarted
% $\ell^p-\ell^q$ minimization methods using generalized Krylov subspaces, 
% Electron. Trans. Numer. Anal., Submitted(2024), pp. 1--24.

% Return defaults values for the options if lplq('defaults') is used
if nargin==1 && isequal(A,'defaults') && nargout==1
    x=struct('p',2,'q',0.1,'L',1,'mu',[],'initialGuess',[],'choiceRule','GCV','majorantType','adaptive',...
        'epsilon',1e-3,'tol',1e-4,'maxIt',100,'noiseNorm',[],'tau',1.01,'size',[],'trainingPercent',90,...
        'trainingTestNum',10,'trainingMu',logspace(-3,1,10),'restart',101,'waitbar','off','xTrue',[]);
    outInfo=[];
    return
end
% Check options structure
if nargin==2
    options=struct('p',2,'q',0.1,'L',1,'mu',[],'initialGuess',[],'choiceRule','GCV','majorantType','adaptive',...
        'epsilon',1e-3,'tol',1e-4,'maxIt',100,'noiseNorm',[],'tau',1.01,'size',[],'trainingPercent',90,...
        'trainingTestNum',10,'trainingMu',logspace(-3,1,10),'restart',101,'waitbar','off','xTrue',[]);
end
if nargin==3
    options=checkOptions(options);
end
% Determine if the outInfo structure needs to be created
if nargout==1
    options.outInfo=0;
else
    options.outInfo=1;
end


% Use the appropriate version of lplq depending on the choice rule for the
% regularization parameter
switch options.choiceRule
    case 'DP'
        [x,outInfo]=lplqNS(A,b,options);
    case 'GCV'
        [x,outInfo]=lplqNS(A,b,options);
    case 'RWP'
        [x,outInfo]=lplqNS(A,b,options);
    case 'CV'
        [x,outInfo]=lplqCV(A,b,options);
    case 'MCV'
        [x,outInfo]=lplqMCV(A,b,options);
    otherwise
        [x,outInfo]=lplqS(A,b,options);
end
end
%% lplq for stationary value of mu 
function [x,outInfo]=lplqS(A,b,options)
% Retrieve informations from the options structure
p=options.p; % Value of p
q=options.q; % Value of q
mu=options.mu; % Value of mu
L=options.L; % Regularization operator L
x0=options.initialGuess; % Initial guess
e=options.epsilon; % Value of the smoothing parameter epsilon
tol=options.tol; % Tollerance for the stopping criterion
maxit=options.maxIt; % Maximum number of iterations
rest=options.restart; % Number of iterations every which to perform a restart
x_true=options.xTrue; % Exact solution
if isequal(options.waitbar,'on') % Determine if a waitbar is required
    wbar=1;
else
    wbar=0;
end
MT=options.majorantType; % Type of majorant
OI=options.outInfo; % Flag for the construction of the outInfo structure
if OI
    options.outInfo=struct('iter',[],'mu',mu,'res',zeros(maxit+1,1),'RRE',[]);
else
    outInfo=[];
end
if wbar % Initialize waitbar
    ww=waitbar(0,'l_p-l_q - Computations in progress...');
end

% Run Majorization Minimization iterations depending on the majorant type
switch MT
    case 'adaptive' % Adaptive majorant
        % Initial guesses
        if isempty(x0)
            x=A'*b;
        else
            x=x0;
        end
        y=norm(x);
        V=x/y;
        AV=A*V;
        LV=L*V;
        v=AV*y-b;
        u=LV*y;
        if OI
            outInfo.res(1)=norm(v);
            if ~isempty(x_true)
                outInfo.RRE=zeros(maxit+1,1);
                outInfo.RRE(1)=norm(x-x_true)/norm(x_true);
            end
        end
        % Run the iterations
        for k=1:maxit
            % Save the previous iteration. If it is not the first iteration
            % and if at the previous step a restart was not performed it
            % has to be padded with a 0 for the coefficient of the new
            % vector in the GKS
            if mod(k,rest)==1
                yold=y;
            else
                yold=[y;0];
            end
            % Weights for approximating the lp and lq norms with the l2
            % norm
            omegaf=(v.^2+e^2).^(p/2-1); % Fidelity term
            omegar=(u.^2+e^2).^(q/2-1); % Regularization term
            % Perform the economic QR factorizations of the projected
            % matrices
            [QA,RA]=qr(omegaf.^0.5.*AV,0);
            [~,RL]=qr(omegar.^0.5.*LV,0);
            % Solve the projected l2-l2 problem
            y=[RA;sqrt(mu)*RL]\[QA'*(omegaf.^0.5.*b);zeros(size(RL,1),1)];
            if wbar
                waitbar(k/(maxit-1))
            end
            % Check the stopping criterion. The method is not stopped at
            % the first iteration and immediately after a restart
            if (norm(y-yold)<=tol*norm(yold) || k==maxit) && mod(k,rest)~=1
                x=V*y;
                if OI
                    outInfo.iter=k;
                    outInfo.res(k+1)=norm(AV*y-b);
                    outInfo.res=outInfo.res(1:k+1);
                    if ~isempty(x_true)
                        outInfo.RRE(k+1)=norm(x-x_true)/norm(x_true);
                        outInfo.RRE=outInfo.RRE(1:k+1);
                    end
                end
                break
            end
            % Compute the new vectors for constructing the new majorants
            v=AV*y-b;
            u=LV*y;
            if OI
                outInfo.res(k+1)=norm(v);
                if ~isempty(x_true)
                    outInfo.RRE(k+1)=norm(V*y-x_true)/norm(x_true);
                end
            end
            % If k is a multiple of rest, the space is restarted. Otherwise
            % the space is enlarged
            if mod(k,rest)==0
                x=V*y;
                clear V
                clear AV
                clear LV
                y=norm(x);
                V=x/y;
                AV=A*V;
                LV=L*V;
            else
                % Compute the normalized and orthogalized residual of the
                % normal equations
                r=A'*(omegaf.*(AV*y-b))+mu*(L'*(omegar.*(LV*y)));
                r=r-V*(V'*r);
                vnew=r/norm(r);
                % Enlarge the space
                W=[V, vnew];
                AW=[AV, A*vnew];
                LW=[LV, L*vnew];
                clear V
                clear AV
                clear LV
                V=W;
                AV=AW;
                LV=LW;
            end
        end
        if k==maxit
            x=V*y;
            if OI
                outInfo.iter=maxit;
            end
        end
    case 'fixed' % Fixed majorant
        % Initial guesses and factorizations
        if isempty(x0)
            x=A'*b;
        else
            x=x0;
        end
        y=norm(x);
        V=x/y;
        AV=A*V;
        RA=norm(AV);
        QA=AV/RA;
        LV=L*V;
        RL=norm(LV);
        QL=LV/RL;
        v=AV*y-b;
        u=LV*y;
        if OI
            outInfo.res(1)=norm(v);
            if ~isempty(x_true)
                outInfo.RRE=zeros(maxit+1,1);
                outInfo.RRE(1)=norm(x-x_true)/norm(x_true);
            end
        end
        % Compute the value of the regularization parameter for the l2-l2
        % problems
        eta=mu*e^((q-p)/2);
        % Run the iterations
        for k=1:maxit
            % Save the previous iteration. If it is not the first iteration
            % and if at the previous step a restart was not performed it
            % has to be padded with a 0 for the coefficient of the new
            % vector in the GKS
            if mod(k,rest)==1
                yold=y;
            else
                yold=[y;0];
            end
            % Compute the shifts for the approximating l2-l2 problems
            omegaf=v.*(1-((v.^2+e^2)/e^2).^(p/2-1)); % Fidelity term
            omegar=u.*(1-((u.^2+e^2)/e^2).^(q/2-1)); % Regularization term
            % Solve the projected problem
            y=[RA;sqrt(eta)*RL]\[QA'*(b+omegaf);sqrt(eta)*(QL'*omegar)];
            if wbar
                waitbar(k/(maxit-1))
            end
            % Check the stopping criterion. The method is not stopped at
            % the first iteration and immediately after a restart
            if (norm(y-yold)<=tol*norm(yold) || k==maxit) && mod(k,rest)~=1
                x=V*y;
                if OI
                    outInfo.iter=k;
                    outInfo.res(k+1)=norm(AV*y-b);
                    outInfo.res=outInfo.res(1:k+1);
                    if ~isempty(x_true)
                        outInfo.RRE(k+1)=norm(x-x_true)/norm(x_true);
                        outInfo.RRE=outInfo.RRE(1:k+1);
                    end
                end
                break
            end
            % Compute the new vectors for constructing the new majorants
            v=AV*y-b;
            u=LV*y;
            if OI
                outInfo.res(k+1)=norm(v);
                if ~isempty(x_true)
                    outInfo.RRE(k+1)=norm(V*y-x_true)/norm(x_true);
                end
            end
            % If k is a multiple of rest, the space is restarted. Otherwise
            % the space is enlarged and the QR factorizations are updated
            if mod(k,rest)==0
                x=V*y;
                clear V
                clear AV
                clear LV
                clear QA
                clear RA
                clear QL
                clear RL
                y=norm(x);
                V=x/y;
                AV=A*V;
                RA=norm(AV);
                QA=AV/RA;
                LV=L*V;
                RL=norm(LV);
                QL=LV/RL;
            else
                [QA,RA,QL,RL,AV,LV,V]=updateQR(A,L,QA,RA,QL,RL,AV,LV,V,eta,b,y,omegaf,omegar);
            end
        end
        if k==maxit
            x=V*y;
            if OI
                outInfo.iter=maxit;
            end
        end
end
% Close the waitbar. The try instruction is used to avoid errors if the
% waitbar was closed by the user before the end of the program.
if wbar
    try
        close(ww)
    catch
        warning('The waitbar was not closed')
    end
end
end
%% lplq for nonstationary value of mu 
function [x,outInfo]=lplqNS(A,b,options)
% Retrieve informations from the options structure
p=options.p; % Value of p
q=options.q; % Value of q
ch=options.choiceRule; % Selection of the choice rule
delta=options.noiseNorm; % Norm of the noise (used by the DP)
tau=options.tau; % Safeguard parameter for the DP
x0=options.initialGuess; % Initial guess
L=options.L; % Regularization operator L
e=options.epsilon; % Smoothing parameter epsilon
tol=options.tol; % Tollerance for the stopping criterion
maxit=options.maxIt; % Maximum number of iterations
rest=options.restart; % Number of iterations after which a restart is performed
sizes=options.size; % Size of the solution
if isempty(sizes)
    sizes=length(b);
end
x_true=options.xTrue; % Exact solution
if isequal(options.waitbar,'on') % Flag for the waitbar
    wbar=1;
else
    wbar=0;
end
MT=options.majorantType; % Majorant type
OI=options.outInfo; % Flag for the construction of the outInfo structure
if OI
    outInfo=struct('iter',[],'mu',zeros(maxit,1),'res',zeros(maxit+1,1),'RRE',[]);
else
    outInfo=[];
end
if wbar
    ww=waitbar(0,'l_p-l_q - Computations in progress...');
end
% Compute a smoothed version of b if the GCV is used with p<=1
if strcmp(ch,'GCV') && p<=1
    nD=numel(sizes); % Check the dimensionality of B
    % Check if b is a vectorized version of a multidimensional vector
    switch nD
        case 1
            % Determine the Gaussian kernel and its center to use to convolve with b
            s=linspace(-2,2,min(5,length(b)));
            center=round(length(s)/2);
            s=exp(-s(:).^2/2);
            s=padarray(s,size(b)-size(s),0,'post'); % Pad the kernel so that it has the same size of b
            s=fft(circshift(s,1-center)); % Compute the Fourier transform of the kernel
            bsmooth=real(ifft(s.*fft(b))); % Convolve the kernel with b (with periodic boundary conditions)
        case 2 % Two dimensions
            B=reshape(b,sizes(:)'); % Reshape B to be a multimensional vector
            % Determine the Gaussian kernel and its center to use to
            % convolve with B
            s=linspace(-2,2,min(5,min(sizes)));
            [X,Y]=meshgrid(s,s);
            K=exp(-(X.^2+Y.^2)/2);
            center=round(size(K)/2);
            K=padarray(K,size(B)-size(K),0,'post'); % Pad the kernel so that it has the same size of B
            K=fft2(circshift(K,1-center)); % Compute the multidimensional Fourier transform of the kernel
            Bsmooth=real(ifft2(K.*fft2(B))); % Convolve the kernel with B (with periodic boundary conditions)
            bsmooth=Bsmooth(:); % Vectorize the smoothed B
        case 3 % Three dimensions
            B=reshape(b,sizes(:)'); % Reshape B to be a multimensional vector
            % Determine the Gaussian kernel and its center to use to
            % convolve with B
            s=linspace(-2,2,min(5,min(sizes)));
            [X,Y,Z]=meshgrid(s,s,s);
            K=exp(-(X.^2+Y.^2+Z.^2)/2);
            center=round(size(K)/2);
            K=padarray(K,size(B)-size(K),0,'post');% Pad the kernel so that it has the same size of B
            K=fftn(circshift(K,1-center)); % Compute the multidimensional Fourier transform of the kernel
            Bsmooth=real(ifftn(K.*fftn(B))); % Convolve the kernel with B (with periodic boundary conditions)
            bsmooth=Bsmooth(:);  % Vectorize the smoothed B
        otherwise
            error('The multidimensional smoothed GCV is implemented only for two and three dimensions.')
    end
elseif strcmp(ch,'RWP')
    if numel(sizes)==1
        sizes=[sizes 1];
    end
end
% Run Majorization Minimization iterations depending on the majorant type
switch MT
    case 'adaptive' % Adaptive majorant
        % Initial guesses and factorizations
        if isempty(x0)
            x=A'*b;
        else
            x=x0;
        end
        y=norm(x);
        V=x/y;
        AV=A*V;
        LV=L*V;
        v=AV*y-b;
        u=LV*y;
        if OI
            outInfo.res(1)=norm(v);
            if ~isempty(x_true)
                outInfo.RRE=zeros(maxit+1,1);
                outInfo.RRE(1)=norm(x-x_true)/norm(x_true);
            end
        end
        % Run the iterations
        for k=1:maxit
            % Save the previous iteration. If it is not the first iteration
            % and if at the previous step a restart was not performed it
            % has to be padded with a 0 for the coefficient of the new
            % vector in the GKS
            if mod(k,rest)==1
                yold=y;
            else
                yold=[y;0];
            end
            % Weights for approximating the lp and lq norms with the l2
            % norm
            omegaf=(v.^2+e^2).^(p/2-1); % Fidelity term 
            omegar=(u.^2+e^2).^(q/2-1); % Regularization term
            % Perform the economic QR factorizations of the projected
            % matrices
            [QA,RA]=qr(omegaf.^0.5.*AV,0);
            [~,RL]=qr(omegar.^0.5.*LV,0);
            % Determine the regularization parameter using the choice rule
            % selected
            bhat=QA'*(omegaf.^0.5.*b); % Projected right-hand side
            switch ch
                case 'RWP' % Residual whiteness principle
                    mu=RWP_adapt(b,AV,RA,RL,QA,omegaf,sizes);
                case 'GCV' % Generalized cross validation
                    % Select if the GCV or the smoothed GCV needs to be
                    % used
                    if p>1
                        mu=GCV(RA,RL,bhat);
                    else
                        bt=QA'*(omegaf.^0.5.*bsmooth);
                        mu=GCV(RA,RL,bt);
                    end
            end
            % Solve the projected l2-l2 problem
            y=[RA;sqrt(mu)*RL]\[bhat;zeros(size(RL,1),1)];
            if wbar
                waitbar(k/(maxit-1))
            end
            % Check the stopping criterion. The method is not stopped at
            % the first iteration and immediately after a restart
            if (norm(y-yold)<=tol*norm(yold) || k==maxit) && mod(k,rest)~=1
                x=V*y;
                if OI
                    outInfo.mu(k)=mu;
                    outInfo.mu=outInfo.mu(1:k);
                    outInfo.iter=k;
                    outInfo.res(k+1)=norm(AV*y-b);
                    outInfo.res=outInfo.res(1:k+1);
                    if ~isempty(x_true)
                        outInfo.RRE(k+1)=norm(x-x_true)/norm(x_true);
                        outInfo.RRE=outInfo.RRE(1:k+1);
                    end
                end
                break
            end
            % Compute the new vectors for constructing the new majorants
            v=AV*y-b;
            u=LV*y;
            if OI
                outInfo.mu(k)=mu;
                outInfo.res(k+1)=norm(v);
                if ~isempty(x_true)
                    outInfo.RRE(k+1)=norm(V*y-x_true)/norm(x_true);
                end
            end
            % If k is a multiple of rest, the space is restarted. Otherwise
            % the space is enlarged 
            if mod(k,rest)==0
                x=V*y;
                clear V
                clear AV
                clear LV
                y=norm(x);
                V=x/y;
                AV=A*V;
                LV=L*V;
            else
                % Compute the normalized and orthogalized residual of the
                % normal equations
                r=A'*(omegaf.*(AV*y-b))+mu*(L'*(omegar.*(LV*y)));
                r=r-V*(V'*r);
                vnew=r/norm(r);
                % Enlarge the space
                W=[V, vnew];
                AW=[AV, A*vnew];
                LW=[LV, L*vnew];
                clear V
                clear AV
                clear LV
                V=W;
                AV=AW;
                LV=LW;
            end
        end
        if k==maxit
            x=V*y;
            if OI
                outInfo.iter=maxit;
            end
        end
    case 'fixed' % Fixed majorant
        % Initial guesses and factorizations
        if isempty(x0)
            x=A'*b;
        else
            x=x0;
        end
        y=norm(x);
        V=x/y;
        AV=A*V;
        RA=norm(AV);
        QA=AV/RA;
        LV=L*V;
        RL=norm(LV);
        QL=LV/RL;
        v=AV*y-b;
        u=LV*y;
        if OI
            outInfo.res(1)=norm(v);
            if ~isempty(x_true)
                outInfo.RRE=zeros(maxit+1,1);
                outInfo.RRE(1)=norm(x-x_true)/norm(x_true);
            end
        end
        % Run the iterations
        for k=1:maxit
            % Save the previous iteration. If it is not the first iteration
            % and if at the previous step a restart was not performed it
            % has to be padded with a 0 for the coefficient of the new
            % vector in the GKS
            if mod(k,rest)==1
                yold=y;
            else
                yold=[y;0];
            end
            % Compute the shifts for the approximating l2-l2 problems
            omegaf=v.*(1-((v.^2+e^2)/e^2).^(p/2-1));
            omegar=u.*(1-((u.^2+e^2)/e^2).^(q/2-1));
            % Determine the regularization parameter using the choice rule
            % selected
            switch ch
                case 'RWP' % Residual whiteness principle
                    eta=RWP_fixed(b,AV,RA,RL,QA,omegaf,QL,omegar,sizes);
                case 'DP' % Discrepancy principle
                    eta=DP(RA,RL,QA,b,QL,omegar,tau,delta);
            end
            % Solve the projected problem
            y=[RA;sqrt(eta)*RL]\[QA'*(b+omegaf);sqrt(eta)*(QL'*omegar)];
            if wbar
                waitbar(k/(maxit-1))
            end
            % Check the stopping criterion. The method is not stopped at
            % the first iteration and immediately after a restart
            if (norm(y-yold)<=tol*norm(yold) || k==maxit) && mod(k,rest)~=1
                x=V*y;
                if OI
                    outInfo.mu(k)=eta*e^((p-q)/2); % The ouput value is selected to be mu and not eta
                    outInfo.mu=outInfo.mu(1:k);
                    outInfo.iter=k;
                    outInfo.res(k+1)=norm(AV*y-b);
                    outInfo.res=outInfo.res(1:k+1);
                    if ~isempty(x_true)
                        outInfo.RRE(k+1)=norm(x-x_true)/norm(x_true);
                        outInfo.RRE=outInfo.RRE(1:k+1);
                    end
                end
                break
            end
            % Compute the new vectors for constructing the new majorants
            v=AV*y-b;
            u=LV*y;
            if OI
                outInfo.mu(k)=eta*e^((p-q)/2); % The ouput value is selected to be mu and not eta
                outInfo.res(k+1)=norm(v);
                if ~isempty(x_true)
                    outInfo.RRE(k+1)=norm(V*y-x_true)/norm(x_true);
                end
            end
            % If k is a multiple of rest, the space is restarted. Otherwise
            % the space is enlarged and the QR factorizations are updated
            if mod(k,rest)==0
                x=V*y;
                clear V
                clear AV
                clear LV
                clear QA
                clear RA
                clear QL
                clear RL
                y=norm(x);
                V=x/y;
                AV=A*V;
                RA=norm(AV);
                QA=AV/RA;
                LV=L*V;
                RL=norm(LV);
                QL=LV/RL;
            else
                [QA,RA,QL,RL,AV,LV,V]=updateQR(A,L,QA,RA,QL,RL,AV,LV,V,eta,b,y,omegaf,omegar);
            end
        end
        if k==maxit
            x=V*y;
            if OI
                outInfo.iter=maxit;
            end
        end
end
% Close the waitbar. The try instruction is used to avoid errors if the
% waitbar was closed by the user before the end of the program.
if wbar
    try
        close(ww)
    catch
        warning('The waitbar was not closed')
    end
end
end
%% lplq for stationary value of mu with selected rows removed 
function x=lplqRemoveRows(A,b,options,J)
% This function implements the same algorithm as lplqS where some rows of A
% and b are removed. The vector J is the same length of b (that is the same
% number of rows of A). If J(i)==1, the i-th row of A is kept, if J(i)==0, 
% the i-th row of A is removed.

% Retrieve informations from the options structure
p=options.p; % Value of p
q=options.q; % Value of q
mu=options.mu; % Value of mu
L=options.L; % Regularization operator L
x0=options.initialGuess; % Initial guess
e=options.epsilon; % Smoothing parameter epsilon
tol=options.tol; % Tollerance for the stopping criterion
maxit=options.maxIt; % Maximum number of iterations
rest=options.restart; % Number of iterations after which a restart is performed
MT=options.majorantType; % Majorant type

% Run Majorization Minimization iterations depending on the majorant type
switch MT
    case 'adaptive' % Adaptive majorant
        % Initial guesses and factorizations
        if isempty(x0)
            x=AT_MvP(A,b,J);
        else
            x=x0;
        end
        y=norm(x);
        V=x/y;
        AV=A_MvP(A,V,J);
        LV=L*V;
        v=AV*y-b;
        u=LV*y;
        % Run the iterations
        for k=1:maxit
            % Save the previous iteration. If it is not the first iteration
            % and if at the previous step a restart was not performed it
            % has to be padded with a 0 for the coefficient of the new
            % vector in the GKS
            if mod(k,rest)==1
                yold=y;
            else
                yold=[y;0];
            end
            % Weights for approximating the lp and lq norms with the l2
            % norm
            omegaf=(v.^2+e^2).^(p/2-1);
            omegar=(u.^2+e^2).^(q/2-1);
            % Perform the economic QR factorizations of the projected
            % matrices
            [QA,RA]=qr(omegaf.^0.5.*AV,0);
            [~,RL]=qr(omegar.^0.5.*LV,0);
            % Solve the projected l2-l2 problem
            y=[RA;sqrt(mu)*RL]\[QA'*(omegaf.^0.5.*b);zeros(size(RL,1),1)];
            % Check the stopping criterion. The method is not stopped at
            % the first iteration and immediately after a restart
            if (norm(y-yold)<=tol*norm(yold) || k==maxit) && mod(k,rest)~=1
                x=V*y;
                break
            end
            % Compute the new vectors for constructing the new majorants
            v=AV*y-b;
            u=LV*y;
            % If k is a multiple of rest, the space is restarted. Otherwise
            % the space is enlarged
            if mod(k,rest)==0
                x=V*y;
                clear V
                clear AV
                clear LV
                y=norm(x);
                V=x/y;
                AV=A_MvP(A,V,J);
                LV=L*V;
            else
                % Compute the normalized and orthogalized residual of the
                % normal equations
                r=AT_MvP(A,(omegaf.*(AV*y-b)),J)+mu*(L'*(omegar.*(LV*y)));
                r=r-V*(V'*r);
                vnew=r/norm(r);
                % Enlarge the space
                W=[V, vnew];
                AW=[AV, A*vnew];
                LW=[LV, L*vnew];
                clear V
                clear AV
                clear LV
                V=W;
                AV=AW;
                LV=LW;
            end
        end
    case 'fixed' % Fixed majorant
        % Initial guesses and factorizations
        if isempty(x0)
            x=AT_MvP(A,b,J);
        else
            x=x0;
        end
        y=norm(x);
        V=x/y;
        AV=A_MvP(A,V,J);
        RA=norm(AV);
        QA=AV/RA;
        LV=L*V;
        RL=norm(LV);
        QL=LV/RL;
        v=AV*y-b;
        u=LV*y;
        % Compute the value of the regularization parameter for the l2-l2
        % problems
        eta=mu*e^((q-p)/2);
        % Run the iterations
        for k=1:maxit
            % Save the previous iteration. If it is not the first iteration
            % and if at the previous step a restart was not performed it
            % has to be padded with a 0 for the coefficient of the new
            % vector in the GKS
            if mod(k,rest)==1
                yold=y;
            else
                yold=[y;0];
            end
            % Compute the shifts for the approximating l2-l2 problems
            omegaf=v.*(1-((v.^2+e^2)/e^2).^(p/2-1)); % Fidelity term
            omegar=u.*(1-((u.^2+e^2)/e^2).^(q/2-1)); % Regularization term
            % Solve the projected problem
            y=[RA;sqrt(eta)*RL]\[QA'*(b+omegaf);sqrt(eta)*(QL'*omegar)];
            % Check the stopping criterion. The method is not stopped at
            % the first iteration and immediately after a restart
            if (norm(y-yold)<=tol*norm(yold) || k==maxit) && mod(k,rest)~=1
                x=V*y;
                break
            end
            % Compute the new vectors for constructing the new majorants
            v=AV*y-b;
            u=LV*y;
            % If k is a multiple of rest, the space is restarted. Otherwise
            % the space is enlarged and the QR factorizations are updated
            if mod(k,rest)==0
                x=V*y;
                clear V
                clear AV
                clear LV
                clear QA
                clear RA
                clear QL
                clear RL
                y=norm(x);
                V=x/y;
                AV=A_MvP(A,V,J);
                RA=norm(AV);
                QA=AV/RA;
                LV=L*V;
                RL=norm(LV);
                QL=LV/RL;
            else
                [QA,RA,QL,RL,AV,LV,V]=updateQR_removedRows(A,L,QA,RA,QL,RL,AV,LV,V,eta,b,y,omegaf,omegar,J);
            end
        end
end
end
function y=A_MvP(A,x,J)
% This function implements the matrix-vector product A*x where the some
% rows of A have been removed. The vector J determines which rows of A are
% removed, size(J)==[size(A,1),1]. If J(i)==0, then the i-th row of A is
% removed, if J(i)==1, then the i-th row of A is kept. 
yt=A*x;
y=yt(J);
end
function x=AT_MvP(A,y,J)
% This function implements the matrix-vector product A'*y where the some
% rows of A have been removed. The vector J determines which rows of A are
% removed, size(J)==[size(A,1),1]. If J(i)==0, then the i-th row of A is
% removed, if J(i)==1, then the i-th row of A is kept. 
yt=zeros(size(J));
yt(J)=y;
x=A'*yt;
end
%% lplq with CV 
function [x,outInfo]=lplqCV(A,b,options)
% Retrieve informations from the options structure
tP=options.trainingPercent; % Percentage of rows of A kept to determine mu
NR=numel(b)-fix(numel(b)*tP/100); % Number of rows of A kept to determine mu
tN=options.trainingTestNum; % Number of different tests to perform to avoid statistical errors
mu=options.trainingMu; % Values of mu to test
x_true=options.xTrue; % Exact solution
OI=options.outInfo; % Flag for the construction of the outInfo structure
if OI
    % For the CV the values of iter, mu, and res are related to the final
    % run. The values of resCV and RRECV are related to the runs with
    % different mus and different training sets
    outInfo=struct('iter',[],'mu',[],'res',[],'RRE',[],'resCV',zeros(length(mu),tN),'RRECV',zeros(length(mu),tN));
else
    outInfo=[];
end
% Each increment of the waitbar corresponds to a complete run of lp-lq
if isequal(options.waitbar,'on')
    wbar=1;
else
    wbar=0;
end
% Preallocation
Res=zeros(size(mu));
muF=0;
if wbar
    ww=waitbar(0,'l_p-l_q - Computations in progress');
end
% Run lp-lq for different values of training sets and mu 
for i=1:tN
    % Construct randomly the training set
    J=ones(size(b));
    P=randperm(numel(b));
    J(P(1:NR))=0;
    Jcomp=1-J;
    Jcomp=logical(Jcomp);
    J=logical(J);
    % Remove the rows from b
    bt=b(J);
    % Run lp-lq for each value of mu that needs to be tested
    for j=1:length(mu)
        options.mu=mu(j);
        x=lplqRemoveRows(A,bt,options,J);
        % Determine the norm of the residual on the discarded rows
        r=b-A*x;
        Res(j)=norm(r(Jcomp));
        if OI
            outInfo.resCV(j,i)=Res(j);
            if ~isempty(x_true)
                outInfo.RRECV(j,i)=norm(x-x_true)/norm(x_true);
            end
        end
        if wbar
            waitbar((length(mu)*(i-1)+j)/(length(mu)*tN))
        end
    end
    % Determine the mu that minimizes the residual on the discarded rows
    [~,idx]=min(Res);
    % Average the mu obtained for each training set
    muF=muF+mu(idx);
end
muF=muF/tN;
% Run lp-lq for the full problem with the determined mu
options.mu=muF;
options.waitbar='off';
[x,outInfoF]=lplqS(A,b,options);
if OI
    outInfo.mu=muF;
    outInfo.iter=outInfoF.iter;
    outInfo.res=outInfoF.res;
    outInfo.RRE=outInfoF.RRE;
    if isempty(x_true)
        outInfo.RRECV=[];
    end
end
% Close the waitbar. The try instruction is used to avoid errors if the
% waitbar was closed by the user before the end of the program.
if wbar
    try
        close(ww)
    catch
        warning('The waitbar was not closed.')
    end
end
end
%% lplq with MCV
function [x,outInfo]=lplqMCV(A,b,options)
% Retrieve informations from the options structure
tP=options.trainingPercent; % Percentage of rows of A kept to determine mu
NR=numel(b)-fix(numel(b)*tP/100); % Number of rows of A kept to determine mu
tN=options.trainingTestNum; % Number of different tests to perform to avoid statistical errors
mu=options.trainingMu; % Values of mu to test
x_true=options.xTrue; % Exact solution
OI=options.outInfo; % Flag for the construction of the outInfo structure
if OI
    % For the MCV the values of iter, mu, and res are related to the final
    % run. The values of diffMCV and RREMCV are related to the runs with
    % different mus and different training sets
    outInfo=struct('iter',[],'mu',[],'res',[],'RRE',[],'diffMCV',zeros(length(mu),tN),'RREMCV',zeros(length(mu),tN,2));
else
    outInfo=[];
end
% Each increment of the waitbar corresponds to two complete runs of lp-lq
if isequal(options.waitbar,'on')
    wbar=1;
else
    wbar=0;
end
% Preallocation
Diff=zeros(size(mu));
muF=0;
if wbar
    ww=waitbar(0,'l_p-l_q - Computations in progress');
end
% Run lp-lq for different values of training sets and mu 
for i=1:tN
    % Construct randomly the two training set
    J1=ones(size(b));
    P=randperm(numel(b));
    J1(P(1:NR))=0;
    J1=logical(J1);
    J2=ones(size(b));
    P=randperm(numel(b));
    J2(P(1:NR))=0;
    J2=logical(J2);
    % Remove the rows from b
    bt1=b(J1);
    bt2=b(J2);
    % Run lp-lq for each value of mu that needs to be tested
    for j=1:length(mu)
        options.mu=mu(j);
        x1=lplqRemoveRows(A,bt1,options,J1);
        x2=lplqRemoveRows(A,bt2,options,J2);
        % Determine the norm of the difference between the outputs obtained 
        % with two runs of lp-lq with different sets of discarded rows
        Diff(j)=norm(x1-x2);
        if OI
            outInfo.diffMCV(j,i)=Diff(j);
            if ~isempty(x_true)
                outInfo.RREMCV(j,i,1)=norm(x1-x_true)/norm(x_true);
                outInfo.RREMCV(j,i,2)=norm(x2-x_true)/norm(x_true);
            end
        end
        if wbar
            waitbar((length(mu)*(i-1)+j)/(length(mu)*tN))
        end
    end
    % Determine mu that minimizes the difference between the outputs 
    % obtained with two runs of lp-lq with different sets of discarded rows
    [~,idx]=min(Diff);
    % Average the mu obtained for each training set
    muF=muF+mu(idx);
end
muF=muF/tN;
options.mu=muF;
% Run lp-lq for the full problem with the determined mu
options.waitbar='off';
[x,outInfoF]=lplqS(A,b,options);
if OI
    outInfo.mu=muF;
    outInfo.iter=outInfoF.iter;
    outInfo.res=outInfoF.res;
    outInfo.RRE=outInfoF.RRE;
    if isempty(x_true)
        outInfo.RRECV=[];
    end
end
if wbar
    try
        close(ww)
    catch
        warning('The waitbar was not closed.')
    end
end
end
%% Update of QR factorizations with selected rows removed 
function [QA,RA,QL,RL,AV,LV,V]=updateQR_removedRows(A,L,QA,RA,QL,RL,AV,LV,V,eta,b,y,omegaf,omegar,J)
r=AT_MvP(A,(AV*y-b-omegaf),J)+eta*(L'*(LV*y-omegar)); % Compute the residual of the normal equation
r=r-V*(V'*r); % Reorthogonalize the residual
vnew=r/norm(r); % Normalize the residual
V=[V vnew]; % Enlarge the space
Avn=A_MvP(A,vnew,J); % Compute A*vnew
AV=[AV, Avn]; % Enlarge AV
Lvn=L*vnew; % Compute L*vnew
LV=[LV, Lvn]; % Enlarge LV
% Update the QR factorization of AV
rA=QA'*Avn;
qA=Avn-QA*rA;
tA=norm(qA);
qtA=qA/tA;
QA=[QA,qtA];
RA=[RA rA;zeros(1,length(rA)) tA];
% Update the QR factorization of LV
rL=QL'*Lvn;
qL=Lvn-QL*rL;
tL=norm(qL);
qtL=qL/tL;
QL=[QL,qtL];
RL=[RL rL;zeros(1,length(rL)) tL];
end
%% Update of QR factorizations 
function [QA,RA,QL,RL,AV,LV,V]=updateQR(A,L,QA,RA,QL,RL,AV,LV,V,eta,b,y,omegaf,omegar)
r=A'*(AV*y-b-omegaf)+eta*(L'*(LV*y-omegar)); % Compute the residual of the normal equation
r=r-V*(V'*r); % Reorthogonalize the residual
vnew=r/norm(r); % Normalize the residual
V=[V vnew]; % Enlarge the space
Avn=A*vnew;
AV=[AV, Avn]; % Enlarge AV
Lvn=L*vnew;
LV=[LV, Lvn]; % Enlarge LV
% Update the QR factorization of AV
rA=QA'*Avn;
qA=Avn-QA*rA;
tA=norm(qA);
qtA=qA/tA;
QA=[QA,qtA];
RA=[RA rA;zeros(1,length(rA)) tA];
% Update the QR factorization of LV
rL=QL'*Lvn;
qL=Lvn-QL*rL;
tL=norm(qL);
qtL=qL/tL;
QL=[QL,qtL];
RL=[RL rL;zeros(1,length(rL)) tL];
end
%% RWP choice rule (adaptive case) 
function mu=RWP_adapt(b,AV,RA,RL,QA,omegaf,sizes)
resFun=@(mu) b-AV*([RA;sqrt(mu)*RL]\[QA'*(omegaf.^0.5.*b);zeros(size(RL,1),1)]); % Function that determines the residual given mu
mu=fminbnd(@(mu) resWhiteFun(mu,resFun,sizes),1e-6,1e2); % Determine the minimizer of the residual whiteness function
end
%% RWP choice rule (fixed case) 
function eta=RWP_fixed(b,AV,RA,RL,QA,omegaf,QL,omegar,sizes)
resFun=@(eta) b-AV*([RA;sqrt(eta)*RL]\[QA'*(b+omegaf);sqrt(eta)*(QL'*omegar)]); % Function that determines the residual given mu
eta=fminbnd(@(eta) resWhiteFun(eta,resFun,sizes),1e-6,1e2); % Determine the minimizer of the residual whiteness function
end
%% residual whiteness function
function w=resWhiteFun(mu,resFun,sizes)
% This function computes the residual whiteness function given mu and the
% function resFun that computes the residual b-A*x fora given mu
r=resFun(mu);
r=reshape(r,sizes(:)');
rhat=fftn(r);
w=norm(rhat(:).^2,'fro')^2/norm(rhat(:),'fro')^4;
end
%% GCV 
function mu=GCV(RA,RL,b)
% This function determines the regularization parameter mu using the GCV
[U,~,~,S,L]=gsvd(RA,RL); % Compute the gsvd of RA and RL
bhat=U'*b; % Project b  
s=diag(S); % Extract the gen. singular values of RA
l=diag(L); % Extract the gen. singular values of RL 
mu=fminbnd(@(mu) GCVfun(mu,s,l,bhat),1e-4,1e2); % Determine mu minimizing the GCV functional
end
function g=GCVfun(mu,s,l,bhat)
% This function computes the GCV functional given mu
num=sum(mu*l.^2.*bhat.^2./(s.^2+mu*l.^2).^2);
den=sum(length(s)-sum(s.^2./(s.^2+mu*l.^2).^2))^2;
g=num/den;
end
%% DP 
function mu=DP(RA,RL,QA,b,QL,omega,tau,delta)
% This function determines the regularization parameter mu using the DP
[U,V,~,S,L]=gsvd(RA,RL); % Compute the gsvd of RA and RL
s=diag(S); % Extract the gen. singular values of RA
l=diag(L); % Extract the gen. singular values of RL 
bhat=QA'*b; % Project b
btilde=U'*bhat; % Project bhat
omegatilde=V'*(QL'*omega); % Project omega
c=norm(b-QA*bhat)^2-tau^2*delta^2; % Determine the constant
beta=0; % Inizialize beta=1/mu to 0
% Run Newton iterations
for k=1:30
    betaold=beta;
    f=sum((l.*s.*omegatilde-l.^2.*btilde).^2.*(beta*s.^2+l.^2).^-2)+c;
    fprime=-2*sum(s.^2.*(l.*s.*omegatilde-l.^2.*btilde).^2.*(beta*s.^2+l.^2).^-3);
    beta=beta-f/fprime;
    if abs(beta-betaold)<1e-6*abs(betaold)
        break
    end
end
% Compute mu
mu=1/beta;
end
%% Check the option structure
function options=checkOptions(options)
% This function check if the fields in the options structure are populated
% and their coherence

% options.p default value
if ~isfield(options,'p')
    options.p=2;
end
% Check that p is a number between 0 (excluded) and 2 (included)
if ~isnumeric(options.p) || options.p<=0||options.p>2
    error('The value of p must be a number between 0 (excluded) and 2 (included).')
end
% options.q default value
if ~isfield(options,'q')
    options.q=0.1;
end
% Check that q is a number between 0 (excluded) and 2 (included)
if ~isnumeric(options.q) || options.q<=0||options.q>2
    error('The value of q must be a number between 0 (excluded) and 2 (included).')
end
% options.L default value
if ~isfield(options,'L')
    options.L=1;
end
% options.mu default value
if ~isfield(options,'mu')
    options.mu=[];
end
% Check that mu is numeric
if ~isnumeric(options.mu)
    error('options.mu must be a number or empty.')
end
% Check that mu, if it is not empty, is a single number and that is greater than 0
if ~isempty(options.mu) && (length(options.mu)>1 || options.mu<=0)
    error('The value of mu must be a number strictly greater than 0.')
end
% options.initialGuess default value
if ~isfield(options,'initialGuess')
    options.initialGuess=[];
end
% Check that options.initialGuess is numeric
if ~isnumeric(options.initialGuess)
    error('options.initialGuess must be a vector or empty.')
end
% options.choiceRule default value
if ~isfield(options,'choiceRule')
    options.choiceRule='GCV';
end
% Check that the options.choiceRule is among the implemented ones
if ~isequal(options.choiceRule,'GCV') && ~isequal(options.choiceRule,'DP')&& ...
        ~isequal(options.choiceRule,'CV')&& ~isequal(options.choiceRule,'MCV')&& ...
        ~isequal(options.choiceRule,'RWP') &&~isempty(options.choiceRule)
    error('Unknown choice rule.')
end
% If choicheRule is set to empty, then it is assigned the value 'none'
if isempty(options.choiceRule)
    options.choiceRule='none';
end
% If a value for mu is given, the choiceRule is set to 'none' so that
% stationary lp-lq with fixed me is used
if ~isempty(options.mu)
    options.choiceRule='none';
end
% Check that, if the choiceRule is the discrepancy principle, p is set to 2
if isequal(options.choiceRule,'DP') && options.p~=2
    error('If the discrepancy principle is used p must be equal to 2.')
end
% Check that, if the choiceRule is the residual whiteness principle, p is
% set to either 2 or 1
if isequal(options.choiceRule,'RWP') && (options.p~=2 && options.p~=1)
    error('If the residual whiteness principle is used p must be equal to either 2 or 1.')
end
% Check that choiceRule and mu are not empty at the same time
if isempty(options.choiceRule) && isempty(options.mu)
    error('options.choiceRule and options.mu cannot be empty at the same time.')
end
% Check that, if a value for majorantType is given, it is among the
% accepted ones
if isfield(options,'majorantType') && (~isequal(options.majorantType,'adaptive') && ~isequal(options.majorantType,'fixed'))
    error('Unknown majorant type.')
end
% options.majorantType default value. If the choiceRule is 'GCV', then it
% is 'adaptive', otherwise it is 'fixed'
if ~isfield(options,'majorantType')
    if strcmp(options.choiceRule,'GCV')
        options.majorantType='adaptive';
    else
        options.majorantType='fixed';
    end
end
% Set the majorantType to 'fixed' if the choiceRule is set to 'DP'
if isequal(options.choiceRule,'DP')
    options.majorantType='fixed';
end
% Set the majorantType to 'adaptive' if the choiceRule is set to 'GCV'
if isequal(options.choiceRule,'GCV')
    options.majorantType='adaptive';
end
% options.epsilon default value
if ~isfield(options,'epsilon') || isempty(options.epsilon) || ~isnumeric(options.epsilon)
    options.epsilon=1e-3;
end
% options.tol default value
if ~isfield(options,'tol') || isempty(options.tol) || ~isnumeric(options.tol)
    options.tol=1e-4;
end
% options.maxIt default value
if ~isfield(options,'maxIt') || isempty(options.maxIt) || ~isnumeric(options.maxIt)
    options.maxIt=100;
end
% options.noiseNorm default value
if ~isfield(options,'noiseNorm')
    options.noiseNorm=[];
end
% Check that, if the discrepancy principle is selected the norm of the
% noise is provided
if isequal(options.choiceRule,'DP') && (isempty(options.noiseNorm) || ~isnumeric(options.noiseNorm))
    error('If the choice rule is the discrepancy principle the field options.noiseNorm must be populated.')
end
% options.tau default value
if ~isfield(options,'tau') || isempty(options.tau) || ~isnumeric(options.tau)
    options.tau=1.01;
end
% options.size default value
if ~isfield(options,'size') || ~isnumeric(options.size)
    options.size=[];
end
% options.trainingPercent default value
if ~isfield(options,'trainingPercent') || isempty(options.trainingPercent) || ~isnumeric(options.trainingPercent)
    options.trainingPercent=90;
end
% options.trainingTestNum default value
if ~isfield(options,'trainingTestNum') || isempty(options.trainingTestNum) || ~isnumeric(options.trainingTestNum)
    options.trainingTestNum=10;
end
% options.trainingMu default value
if ~isfield(options,'trainingMu') || isempty(options.trainingMu) || ~isnumeric(options.trainingMu)
    options.trainingMu=logspace(-3,1,10);
end
% options.restart default value
if ~isfield(options,'restart') || isempty(options.restart) || ~isnumeric(options.restart)
    options.restart=options.maxIt+1;
end
% options.waitbar default value
if ~isfield(options,'waitbar')
    options.waitbar='off';
end
% If options.waibar is not set to 'on', then it is set automatically to
% 'off'
if ~isequal(options.waitbar,'on')
    options.waitbar='off';
end
% options.xTrue default value
if ~isfield(options,'xTrue') || ~isnumeric(options.xTrue)
    options.xTrue=[];
end
end